﻿using System;
using System.Collections.Generic;
using System.IO;

using PSO2SERVER.Models;
using PSO2SERVER.Object;
using PSO2SERVER.Protocol;
using PSO2SERVER.Protocol.Packets;
using static PSO2SERVER.Zone.Map;

namespace PSO2SERVER.Zone
{
    public struct PSOLocation
    {
        public float RotX, RotY, RotZ, RotW, PosX, PosY, PosZ; // RotX, RotY, RotZ, and RotW make up a Quaternion

        public PSOLocation(float RotX, float RotY, float RotZ, float RotW, float PosX, float PosY, float PosZ)
        {
            this.RotX = RotX;
            this.RotY = RotY;
            this.RotZ = RotZ;
            this.RotW = RotW;

            this.PosX = PosX;
            this.PosY = PosY;
            this.PosZ = PosZ;
        }

        public override string ToString()
        {
            return String.Format("Rot: ({0}, {1}, {2}, {3}) Loc: ({4}, {5}, {6})"
                , RotX, RotY, RotZ, RotW, PosX, PosY, PosZ);
        }
    }

    public class Map
    {
        public string Name { get; set; }
        public MapType Type { get; set; }
        public PSOObject[] Objects { get; set; }
        public PSONPC[] NPCs { get; set; }
        public ObjectHeader MapHeader { get; set; }

        public int MapID { get; set; }
        public int VariantID { get; set; }

        public MapFlags Flags { get; set; }

        public List<Client> Clients { get; set; }

        public GenParam GenerationArgs { get; set; }

        public string InstanceName { get; set; }

        public enum MapType : int
        {
            Lobby = 4,
            Casino = 11,
            MyRoom = 8,
            TeamRoom = 7,
            ChallengeLobby = 5,
            Campship = 1,
            Quest = 0,
            Other = ~0
        };

        [Flags]
        public enum MapFlags : uint
        {
            None = 0,
            MultiPartyArea = 1,
            Unknown1 = 2,
            EnableMap = 4
        }

        public Map(string name, int id, int variant, MapType type, MapFlags flags)
        {
            Name = name;
            MapID = id;
            VariantID = variant;
            Type = type;
            Flags = flags;
            Clients = new List<Client>();
            GenerationArgs = new GenParam();

            Objects = ObjectManager.Instance.GetObjectsForZone(Name);
            NPCs = ObjectManager.Instance.GetNpcSForZone(Name);
        }


        public PSOLocation GetDefaultLocation()
        {
            PSOLocation location;

            switch (Type)
            {
                case MapType.Lobby:
                    location = new PSOLocation(0f, 1f, 0f, 0f, -0.417969f, 0f, 137.375f);
                    break;

                case MapType.Casino:
                    location = new PSOLocation(0f, 1f, 0f, 0f, 2f, 6f, 102f);
                    break;

                default:
                    location = new PSOLocation(0f, 1f, 0f, 0f, 0f, 0f, 0f);
                    break;
            }

            return location;
        }

        /// <summary>
        /// Spawns a client into a map at a given location
        /// </summary>
        /// <param name="c">Client to spawn into map</param>
        /// <param name="location">Location to spawn client at</param>
        /// <param name="questOveride">If this also sets the quest data, specify the quest name here to load the spawn from the bin rather then generate it.</param>
        public void SpawnClient(Client c, PSOLocation location, string questOveride = "")
        {
            if (Clients.Contains(c))
            {
                return;
            }

            // Set area
            if (questOveride != "") // TODO: This is a temporary hack, fix me!!
            {
                //var setAreaPacket = File.ReadAllBytes("Resources\\quests\\" + questOveride + ".bin");
                //c.SendPacket(0x03, 0x24, 0x04, setAreaPacket);

                c.SendPacket(new LoadingLevelPacket(questOveride));
            }
            else
            {
                var _map = new Map("", MapID, VariantID, Type, Flags);
                _map.GenerationArgs.seed = GenerationArgs.seed;
                _map.GenerationArgs.xsize = GenerationArgs.xsize;
                _map.GenerationArgs.ysize = GenerationArgs.ysize;

                c.SendPacket(new MapTransferPacket(_map, c._account.AccountId));
            }

            if (c.CurrentZone != null)
            {
                c.CurrentZone.RemoveClient(c);
            }

            // 设置客户端的账户ID
            c.SendPacket(new SetAccountIDPacket(c._account.AccountId));

            // Spawn Character
            c.SendPacket(new CharacterSpawnPacket(c.Character, location, true, true));
            c.CurrentLocation = location;
            c.CurrentZone = this;

            // Objects
            foreach (PSOObject obj in Objects)
            {
                c.SendPacket(new ObjectSpawnPacket(obj));
            }

            // NPCs
            foreach (PSONPC npc in NPCs)
            {
                c.SendPacket(new NPCSpawnPacket(npc));
            }

            // Spawn for others, Spawn others for me
            CharacterSpawnPacket spawnMe = new CharacterSpawnPacket(c.Character, location, false, true);
            foreach (Client other in Clients)
            {
                other.SendPacket(spawnMe);
                c.SendPacket(new CharacterSpawnPacket(other.Character, other.CurrentLocation, false, true));
            }

            Clients.Add(c);

            Logger.Write("[MAP] {0} 已生成至 {1}.", c._account.Username, Name);

            if (InstanceName != null && ZoneManager.Instance.playerCounter.ContainsKey(InstanceName))
            {
                ZoneManager.Instance.playerCounter[InstanceName] += 1;
            }

        }

        public void RemoveClient(Client c)
        {
            if (!Clients.Contains(c))
                return;

            c.CurrentZone = null;
            Clients.Remove(c);

            foreach (Client other in Clients)
            {
                other.SendPacket(new DespawnPlayerPacket(other._account.AccountId, c._account.AccountId));
            }

            if (InstanceName != null && ZoneManager.Instance.playerCounter.ContainsKey(InstanceName))
            {
                ZoneManager.Instance.playerCounter[InstanceName] -= 1;
                if (ZoneManager.Instance.playerCounter[InstanceName] <= 0)
                {
                    ZoneManager.Instance.playerCounter.Remove(InstanceName);
                    ZoneManager.Instance.instances.Remove(InstanceName);
                }
            }
        }

        public class GenParam
        {
            public GenParam()
            {
            }

            public GenParam(uint seed, uint x, uint y)
            {
                this.seed = seed;
                this.xsize = x;
                this.ysize = y;
            }
            public uint seed, xsize, ysize;
        }
    }

    public struct ZoneSettings
    {
        public uint WorldId;
        public uint Unk1;
        public uint ZoneId;
        /// Map layout id.
        public uint MapId;
        public uint ZoneType;
        public uint Seed;
        public uint Args;
        public uint SizeX;
        public uint SizeY;
        public uint Unk2;
        public uint AreaIndex;
        public uint SubArea;
        public uint Unk3;

        // 构造函数
        public ZoneSettings(uint worldId, uint unk1, uint zoneId, uint mapId, uint zoneType, uint seed,
                            uint args, uint sizeX, uint sizeY, uint unk2, uint areaIndex, uint subArea, uint unk3)
        {
            WorldId = worldId;
            Unk1 = unk1;
            ZoneId = zoneId;
            MapId = mapId;
            ZoneType = zoneType;
            Seed = seed;
            Args = args;
            SizeX = sizeX;
            SizeY = sizeY;
            Unk2 = unk2;
            AreaIndex = areaIndex;
            SubArea = subArea;
            Unk3 = unk3;
        }

        // 重写 ToString 方法（可选）
        public override string ToString()
        {
            return $"WorldId: {WorldId}, ZoneId: {ZoneId}, MapId: {MapId}, SizeX: {SizeX}, SizeY: {SizeY}";
        }
    }

    public class EnemySpawn
    {
        public string EnemyName { get; set; }
        public uint SpawnCategory { get; set; }

        // 默认构造函数
        public EnemySpawn()
        {
            EnemyName = string.Empty;
            SpawnCategory = 0;
        }

        // 带参构造函数
        public EnemySpawn(string enemyName, uint spawnCategory)
        {
            EnemyName = enemyName;
            SpawnCategory = spawnCategory;
        }
    }

    public class ZoneData
    {
        public string Name { get; set; }
        public bool IsSpecialZone { get; set; }
        public uint ZoneId { get; set; }
        public ZoneSettings Settings { get; set; }
        public PSOLocation DefaultLocation { get; set; }
        public List<EnemySpawn> Enemies { get; set; }
        public List<ZoneChunk> Chunks { get; set; }

        // 默认构造函数
        public ZoneData()
        {
            Name = string.Empty;
            IsSpecialZone = false;
            ZoneId = new uint();
            Settings = new ZoneSettings();
            DefaultLocation = new PSOLocation();
            Enemies = new List<EnemySpawn>();
            Chunks = new List<ZoneChunk>();
        }
    }

    public class ZoneChunk
    {
        public uint ZoneId { get; set; }
        public uint ChunkId { get; set; }
        public EnemySpawnType EnemySpawnType { get; set; }
        public List<PSOLocation> EnemySpawnPoints { get; set; }

        // 默认构造函数
        public ZoneChunk()
        {
            ZoneId = new uint();
            ChunkId = 0;
            EnemySpawnType = EnemySpawnType.Disabled;
            EnemySpawnPoints = new List<PSOLocation>();
        }
    }

    public enum EnemySpawnType
    {
        Disabled,
        Automatic,
        AutomaticWithRespawn,
        Manual
    }

    public class AutomaticEnemySpawn
    {
        public uint Min { get; set; }
        public uint Max { get; set; }

        public AutomaticEnemySpawn() { }
        public AutomaticEnemySpawn(uint min, uint max)
        {
            Min = min;
            Max = max;
        }
    }

    public class AutomaticWithRespawnEnemySpawn : AutomaticEnemySpawn
    {
        public TimeSpan RespawnTime { get; set; }

        public AutomaticWithRespawnEnemySpawn() { }
        public AutomaticWithRespawnEnemySpawn(uint min, uint max, TimeSpan respawnTime)
            : base(min, max)
        {
            RespawnTime = respawnTime;
        }
    }
}